// WideMargin. Simple fast bible software.
// Copyright (C) 2011-2013  Daniel Hughes
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.

using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using WideMargin.Database;
using WideMargin.MVCInterfaces;
using WideMargin.Utilities;
using SharpSword;
using WideMargin.Markup;

namespace WideMargin.Controller
{
	/// <summary>
	/// Controller for a view capable of displaying
	/// search results, bible passages and reading plans.
	/// Also allowing search entry, passage navigation and 
	/// full navigation histroy.
	/// </summary>
	public class BibleController : IDisposable
	{
		IBibleView _bibleView;
		QueryQueue _queryQueue;
		BibleBarDecoder _bibleBarDecoder;
		BibleDatabase _bible;
		IFilteredView _searchResultView;
		IPassageView _passageView;
		IGenericView _genericView;
		BookList _bookList;
		Book _currentBook;
		Chapter _currentChapter;
		int _currentVerse;
		EventHandler<VerseEventArgs> _newPassagePageRequest;
		EventHandler<RequestNewResultsPageEventArgs> _newResultPageRequest;
		EventLocker _eventLocker = new EventLocker();
		SearchPageResult _lastShownResult;
		History _history;
		BooksController _booksController;

		/// <summary>
		/// Constructor
		/// </summary>
		/// <param name="bibleView">View for this controller</param>
		/// <param name="queryQueue">Thread safe queue for performing queries</param>
		/// <param name="bibleBarDecoder">Decodes bible bar text</param>
		/// <param name="bible">the bible database</param>
		public BibleController(IBibleView bibleView, QueryQueue queryQueue, BibleBarDecoder bibleBarDecoder, BibleDatabase bible)
		{
			_bibleView = bibleView;
			_bibleView.NavigateForwards += HandleBibleViewNavigateForwards;
			_bibleView.NavigateBackwards += HandleBibleViewNavigateBackwards;
			_bibleView.RequestAboutInfo += HandleBibleViewRequestAboutInfo;

			//setup search veiw
			_searchResultView = _bibleView.SearchView;
			_searchResultView.VerseLinkClicked += HandleVerseLinkClicked;
			_searchResultView.MoreResultsRequired += HandleSearchResultViewMoreResultsRequired;
			_searchResultView.RequestVersification += HandleRequestVersification;

			//setup passage view
			_passageView = _bibleView.PassageView;
			_passageView.RequestNext += HandlePassageViewRequestNext;
			_passageView.RequestPrevious += HandlePassageViewRequestPrevious;
			SetupGenericView(_passageView);

			//setup generic view
			_genericView = _bibleView.GenericView;
			SetupGenericView(_genericView);

			//setup books view
			_booksController = new BooksController(action => ChangeView(action), _bibleView, bible);
			_booksController.VerseLinkClicked += HandleVerseLinkClicked;
			_booksController.BookLinkClicked += HandleBookLinkClicked;

			_bibleView.SearchRequest += HandleBibleViewSearchRequest;
			_bibleView.RequestAutoCompleteList += HandleBibleViewRequestAutoCompleteList;
			_bibleView.BooksView.RequestVersification += HandleRequestVersification;

			_queryQueue = queryQueue;
			_bibleBarDecoder = bibleBarDecoder;
			_bible = bible;
			_bookList = BookList.Get();
			_history = new History();
		}

		void SetupGenericView(IGenericView view)
		{
			view.VerseLinkClicked += HandleVerseLinkClicked;
			view.ActionLinkClicked += HandleActionLinkClicked;
			view.RequestVersification += HandleRequestVersification;
		}

		public void ShowInstallingPage()
		{
			ChangeView(() =>
			{
				StringBuilder builder = new StringBuilder();
				builder.Append("<Text>");
				builder.Append("<Center>");


				builder.Append("<Size value='24'>");
				builder.AppendLine();
				builder.AppendLine("Installing ESV bible");
				builder.AppendLine("Please wait");
				builder.Append("</Size>");

				builder.AppendLine(string.Empty);

				builder.AppendLine("</Center>");
				builder.AppendLine("</Text>");

				_genericView.Text = builder.ToString();

				_bibleView.ChangeTitle("Installing...");

				_bibleView.ShowGenericView();
				_bibleView.Editable = false;
			});
		}

		void HandleRequestVersification(object sender, RequestVersificationEventArgs e)
		{
			e.Versification = _bible.GetVersification();
		}

		/// <summary>
		/// Handles the book link clicked.
		/// </summary>
		/// <param name='sender'>Sender.</param>
		/// <param name='e'>
		/// contains details of the bookthat was clicked and the button which was used.
		/// </param>
		void HandleBookLinkClicked(object sender, LinkClickedEventArgs<string> e)
		{
			NewBooksPageRequest.Fire(this, e);
		}

		/// <summary>
		/// An action link has been clicked.
		/// Performs the clicked action.
		/// </summary>
		void HandleActionLinkClicked(object sender, LinkClickedEventArgs<string> e)
		{
			if (e.Button == 1)
			{
				ShowBooksList();
				return;
			}

			if (e.Button == 2)
			{
				NewBooksPageRequest.Fire(this, e);
				return;
			}
		}

		/// <summary>
		/// Bible view is requesting about info.
		/// </summary>
		/// <param name="sender"></param>
		/// <param name="e"></param>
		void HandleBibleViewRequestAboutInfo(object sender, EventArgs e)
		{
			var aboutInfo = AboutInfo.Get().GetAboutInfo();
			ChangeView(() =>
			{
				_genericView.Text = aboutInfo;

				_bibleView.ChangeTitle("About");

				_bibleView.ShowGenericView();
				_bibleView.Editable = true;
			});
		}

		/// <summary>
		/// Shows the readings for today.
		/// </summary>
		public void ShowDaylyReading()
		{
			var dailyReadings = new DailyReadings(_bible);
			string readings = dailyReadings.GetDailyReadingsHtml();

			ChangeView(() =>
			{
				_genericView.Text = readings;

				_bibleView.ChangeTitle("Todays Readings");

				_bibleView.ShowGenericView();
				_bibleView.Editable = true;
			});
		}

		/// <summary>
		/// Occurs when the view requests to navigate backwards
		/// </summary>
		void HandleBibleViewNavigateBackwards(object sender, EventArgs e)
		{
			_history.Back();
		}

		/// <summary>
		/// Occurs when the view requests to navigate forwards.
		/// </summary>
		private void HandleBibleViewNavigateForwards(object sender, EventArgs e)
		{
			_history.Forward();
		}

		/// <summary>
		/// View is requesting an auto completion list of book names.
		/// </summary>
		void HandleBibleViewRequestAutoCompleteList(object sender, EventArgs e)
		{
			_bibleView.ShowAutoCompleteList(_bookList.AllBookNames());
		}

		/// <summary>
		/// Search Results view is requesting a new search page
		/// </summary>
		void HandleSearchResultViewMoreResultsRequired(object sender, EventArgs e)
		{
			string searchText = _lastShownResult.SearchText;
			int pageSize = _lastShownResult.PageSize;
			var resultsCount = _lastShownResult.Matches;

			Func<SearchPageResult> searchAction = null;
			if (_lastShownResult is SearchPageResultFiltered)
			{
				string book = ((SearchPageResultFiltered)_lastShownResult).Book;
				searchAction = () => _bible.Search(searchText, book, resultsCount, pageSize, _lastShownResult.PageNumber + 1);
			}
			else
			{
				searchAction = () => _bible.Search(searchText, resultsCount, pageSize, _lastShownResult.PageNumber + 1, null);
			}
			_queryQueue.Queue(searchAction,
							  verses => ShowSearchResults(verses, true, false));
			return;
		}

		/// <summary>
		/// Verse link has been clicked
		/// </summary>
		void HandleVerseLinkClicked(object sender, LinkClickedEventArgs<IVerseIdentifier> e)
		{
			if (e.Button == 1)
			{
				//left click open in current
				Book book = _bookList.GetBook(e.LinkTarget.BookName);
				Chapter chapter = book.GetChapter(e.LinkTarget.ChapterNumber);
				NavigatePassage(book, chapter, e.LinkTarget.VerseNumber);
				return;
			}

			if (e.Button == 2)
			{
				//middle click open in new
				_newPassagePageRequest.Fire(this, new VerseEventArgs(e.LinkTarget));
				return;
			}
		}

		/// <summary>
		/// Shows the books list.
		/// </summary>
		/// <param name='book'>
		/// Show with this book's chapters shown.
		/// </param>
		public void ShowBooksList(string book)
		{
			_booksController.ShowChapters(book);
		}

		/// <summary>
		/// Shows the books list with all books shown
		/// </summary>
		public void ShowBooksList()
		{
			_booksController.ShowBookList(_bookList.AllBookNames());
		}

		/// <summary>
		/// Occurs to request a new passage page
		/// </summary>
		public event EventHandler<VerseEventArgs> NewPassagePageRequest
		{
			add
			{
				_eventLocker.Add(ref _newPassagePageRequest, value);
			}
			remove
			{
				_eventLocker.Remove(ref _newPassagePageRequest, value);
			}
		}

		/// <summary>
		/// Occurs to requesta new search results page
		/// </summary>
		public event EventHandler<RequestNewResultsPageEventArgs> NewResultPageRequest
		{
			add
			{
				_eventLocker.Add(ref _newResultPageRequest, value);
			}
			remove
			{
				_eventLocker.Remove(ref _newResultPageRequest, value);
			}
		}

		/// <summary>
		/// fired to request a new books page.
		/// </summary>
		public event EventHandler<LinkClickedEventArgs<string>> NewBooksPageRequest;

		/// <summary>
		/// Handles request to see the previous passage.
		/// </summary>
		void HandlePassageViewRequestPrevious(object sender, EventArgs e)
		{
			Chapter prevChapter = _currentBook.GetPreviousChapter(_currentChapter);
			Book prevBook = _currentBook;
			if (prevChapter == null)
			{
				prevBook = _bookList.PreviousBook(_currentBook);
				if (prevBook == null)
				{
					return;
				}
				prevChapter = prevBook.LastChapter;
			}


			NavigatePassage(prevBook, prevChapter);
		}

		/// <summary>
		/// Handles request to see the next passage.
		/// </summary>
		void HandlePassageViewRequestNext(object sender, EventArgs e)
		{
			Chapter nextChapter = _currentBook.GetNextChapter(_currentChapter);
			Book nextBook = _currentBook;
			if (nextChapter == null)
			{
				nextBook = _bookList.NextBook(_currentBook);
				if (nextBook == null)
				{
					return;
				}
				nextChapter = nextBook.GetChapter(1);
			}

			NavigatePassage(nextBook, nextChapter);
		}

		/// <summary>
		/// Shows the passage containing the specified verse
		/// </summary>
		/// <param name="verse">verse contained in the passage to be shown.</param>
		public void ShowPassage(IVerseIdentifier verse)
		{
			Book book = _bookList.GetBook(verse.BookName);
			NavigatePassage(book, book.GetChapter(verse.ChapterNumber), verse.VerseNumber);
		}

		/// <summary>
		/// Navigates to a passage
		/// </summary>
		/// <param name="book">book to navigate to.</param>
		/// <param name="chapter">chapter of book to navigate to.</param>
		void NavigatePassage(Book book, Chapter chapter)
		{
			NavigatePassage(book, chapter, 1);
		}

		/// <summary>
		/// Navigates to a passage.
		/// </summary>
		/// <param name='book'>Book.</param>
		/// <param name='chapter'>Chapter.</param>
		/// <param name='verse'>Verse.</param>
		void NavigatePassage(Book book, Chapter chapter, int verse)
		{
			_queryQueue.Queue(
				() => _bible.GetChapter(book.Name, chapter.Number), 
				passage => ChangePassage(passage, verse));
		}

		/// <summary>
		/// Bible view is requesting a search be done.
		/// </summary>
		/// <param name="e">Contains search text</param>
		void HandleBibleViewSearchRequest(object sender, StringEventArgs e)
		{
			DoQuery(e.Text, false);
		}

		void DoQuery(string text, bool keepFilters)
		{
			if (text.Contains("\"") && text.Count(charactor => charactor == '\"') % 2 != 0)
			{
				text = text + "\"";
			}

			BibleBarResult result = _bibleBarDecoder.Decode(text, _bible.GetVersification());

			if (result is SearchResult)
			{

				//Was a search query
				if (text.Length  == 0)
				{
					ShowDaylyReading();
					return;
				}
				SearchResult searchResult = (SearchResult)result;
				_queryQueue.Queue(() => searchResult.GetFirstPage(_bible, _bible.GetVersification()),
								  verses => ShowSearchResults(verses, false, keepFilters));
				return;
			}
		}

		/// <summary>
		/// Changes to the specified passage
		/// </summary>
		/// <param name="passage">verses making up the passage</param>
		private void ChangePassage(IEnumerable<IVerse> passage, int verse)
		{
			if (passage.IsEmpty())
			{
				return;
			}

			IVerse firstVerse = passage.First();
			Book newBook = BookList.Get().GetBook(firstVerse.Identifier.BookName);
			Chapter newChapter = newBook.GetChapter(firstVerse.Identifier.ChapterNumber);

			if (_currentBook == newBook &&
			   _currentChapter == newChapter &&
			   _currentVerse == verse)
			{
				ChangeView(() =>
				{
					_bibleView.ShowPassageView();
				});
				return;
			}
			ChangeView(() =>
			{
				_currentBook = newBook;
				_currentChapter = newChapter;
				_currentVerse = verse;

				_bibleView.ChangeTitle(string.Format("{0} {1}", newBook.Name, newChapter.Number));

				_bibleView.ShowPassageView();
				_passageView.ShowVerses(passage);

				if (verse != 1)
				{
					IVerse verseToScrollTo = passage.Where(passageVerse => passageVerse.Identifier.VerseNumber == verse)
												.FirstOrDefault();
					if (verseToScrollTo == null)
					{
						verseToScrollTo = passage.Last();
					}
					_passageView.ScrollToVerse(verseToScrollTo);
				}
			});
		}

		/// <summary>
		/// Change the shown search result.
		/// </summary>
		/// <param name="searchResults">results of search</param>
		/// <param name="keepFilters">true to keep any displayed filters</param>
		void ShowSearchResults(SearchPageResult searchResults, bool append, bool keepFilters)
		{
			if(searchResults.Verses.IsEmpty() && append)
			{
				return;	
			}
			if (searchResults.Verses.IsEmpty())
			{
				if(searchResults.BookResult != null)
				{
					ShowPassage(searchResults.BookResult);	
					return;
				}
				_genericView.Text = new NoResults().GetNoResultPage();
				_bibleView.ShowGenericView();
				return;
			}

			if (append)
			{
				_lastShownResult = searchResults;
				_currentBook = null;
				_currentChapter = null;
				AppendResultsToSearchView(searchResults);
				return;
			}

			IEnumerable<Filter> filters = null;
			if (keepFilters)
			{
				filters = _searchResultView.Filters;
			}
			else
			{
				var matches = searchResults.Matches;
				var bookList = BookList.Get();
				filters =
					from book in matches.Keys
					orderby bookList.GetBook(book).Index
					let queryText = string.Format("{0} in:{1}", searchResults.SearchText, book)
					select new Filter(string.Format("{0}({1})", book, matches[book]), () => DoQuery(queryText, true));
				var filterAll = new Filter(string.Format("All({0})", matches.Values.Sum()), () => DoQuery(searchResults.SearchText, true));
				filters = new Filter[] { filterAll }.Concat(filters);
			}

			var versesHtml = new WideMargin.Database.Markup
				.SearchResult()
				.GetVersesHtml(searchResults.Verses, searchResults.SearchText, searchResults.BookResult);

			//only a new search needs history, append does not because its trigered by
			//scrolling
			ChangeView(() =>
			{
				_lastShownResult = searchResults;
				_currentBook = null;
				_currentChapter = null;
				_searchResultView.SetResults(versesHtml, filters);
				_bibleView.ChangeTitle("Search");
				_bibleView.ShowSearchView();
			});
		}

		/// <summary>
		/// Appends the results to search view without recording in history.
		/// </summary>
		private void AppendResultsToSearchView(SearchPageResult searchResults)
		{
			_lastShownResult = searchResults;
			_currentBook = null;
			_currentChapter = null;
			var searchResultHtml = new Database.Markup
				.SearchResult()
				.GetVersesFragment(searchResults.Verses, searchResults.SearchText);
			_searchResultView.AppendText(searchResultHtml);
			_bibleView.ShowSearchView();
			return;
		}

		/// <summary>
		/// Changes the currently shown view by
		/// running the passed in action.
		/// Also records the action in the history, to
		/// enable history navigtion.
		/// </summary>
		/// <param name="show">Action which changes the view</param>
		public void ChangeView(Action show)
		{
			_history.Perform(show);
		}

		/// <summary>
		/// Clean up unmanaged resources.
		/// </summary>
		public void Dispose()
		{
			//nothing to do yet
		}
	}
}

