// WideMargin. Simple fast bible software.
// Copyright (C) 2011-2013  Daniel Hughes
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.

using System;
using Gtk;
using WideMargin.MVCInterfaces;
using WideMargin.Utilities;
using SharpSword;

namespace WideMargin.GUI
{
	/// <summary>
	/// Class for displaying generic text with verse links.
	/// </summary>
	[System.ComponentModel.ToolboxItem(true)]
	public partial class GenericView : Gtk.Bin, IGenericView
	{
		RichTextView _richTextView;
		
		/// <summary>
		/// Default constructor
		/// </summary>
		public GenericView ()
		{
			this.Build ();
			_richTextView = new RichTextView();
			_richTextView.VerseLinkClicked += HandleTagLinkClicked;
			_richTextView.ActionLinkClicked += HandleActionLinkClicked;
			_richTextView.RequestVersification += (sender, e) => RequestVersification.Fire(this, e);
			_richTextView.ScrollbarAtBottom += (sender, e) => ScrollbarAtBottom.Fire (this, e);
				
			Add(_richTextView);
		}
		
		public bool VScrollbarVisible
		{
			get
			{
				return _richTextView.VScrollbarVisible;	
			}
		}
		
		/// <summary>
		/// An action link has been clicked, pass on the event
		/// </summary>
		void HandleActionLinkClicked (object sender, LinkClickedEventArgs<string> e)
		{
			if(e.Button == 3)
			{
				Menu menu = new Menu();
				MenuItem aboutMenuItem = new MenuItem("Open Link in New Tab");
				aboutMenuItem.Activated += (sender1, args) =>
				{
					var eventArgs = new LinkClickedEventArgs<string>(e.LinkTarget, 2);
					ActionLinkClicked.Fire(this, eventArgs);
				};
				menu.Add(aboutMenuItem);
				
				menu.ShowAll();
				menu.Popup(null, null, MenuPosition, 0, 0);
				return;
			}
			ActionLinkClicked.Fire(this, e);
		}
		
		/// <summary>
		/// Text including markup
		/// </summary>
		public string Text
		{
			set
			{
				this.Invoke (() => 
				{
					_richTextView.Text = value;
				});
			}
		}
		
		public void AppendText(string text)
		{
			this.Invoke (() => 
			{
				_richTextView.AppendText(text);
			});
		}

		/// <summary>
		/// A verse link has been cliked, Passes the event up to the controller.
		/// </summary>
		void HandleTagLinkClicked (object sender, LinkClickedEventArgs<IVerseIdentifier> e)
		{
			if(e.Button == 3)
			{
				Menu menu = new Menu();
				MenuItem aboutMenuItem = new MenuItem("Open Link in New Tab");
				aboutMenuItem.Activated += (sender1, args) =>
				{
					var eventArgs = new LinkClickedEventArgs<IVerseIdentifier>(e.LinkTarget, 2);
					VerseLinkClicked.Fire(this, eventArgs);
				};
				menu.Add(aboutMenuItem);
				
				menu.ShowAll();
				menu.Popup(null, null, MenuPosition, 0, 0);
				return;
			}
			VerseLinkClicked.Fire(this, e);
		}
		
		/// <summary>
		/// Callback for the setting the position of the rightclick menu
		/// </summary>
		void MenuPosition(Menu menu, out int x, out int y, out bool pushIn)
	    {
			int parentX;
			int parentY;
			
			ParentWindow.GetOrigin(out parentX, out parentY);
			Gdk.ModifierType modifierType;
			ParentWindow.GetPointer(out x, out y, out modifierType);
			x += parentX;
			y += parentY;
			pushIn = true;
		}
		
		public void ExecuteScript(string text)
		{
			_richTextView.ExecuteScript(text);	
		}
		
		/// <summary>
		/// Occurs when a verse link is clicked.
		/// </summary>
		public event EventHandler<LinkClickedEventArgs<IVerseIdentifier>> VerseLinkClicked;
		
		/// <summary>
		/// Occurs when a action link is clicked.
		/// </summary>
		public event EventHandler<LinkClickedEventArgs<string>> ActionLinkClicked;
		
		public event EventHandler<EventArgs> ScrollbarAtBottom;
		
		public event EventHandler<RequestVersificationEventArgs> RequestVersification;
	}
}

