// WideMargin. Simple fast bible software.
// Copyright (C) 2011  Daniel Hughes
// 
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
// 
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.

using System;
using System.Linq;
using Gtk;
using WideMargin.MVCInterfaces;
using System.Collections.Generic;
using WideMargin.Utilities;
using SharpSword;
using WebKit;
using System.Threading;
using System.Threading.Tasks;

namespace WideMargin.GUI
{
	/// <summary>
	/// TextView which supports basic XML style markup.
	/// </summary>
	[System.ComponentModel.ToolboxItem(true)]
	public partial class RichTextView : Gtk.Bin
	{
		WebView _textView;
		GaplessScrolledWindow _scrolledWindow;
		System.Action _adjustmentChangedAction;
		ManualResetEvent _waitForLoad;
		volatile bool _loading = false;
		
		/// <summary>
		/// Constructor
		/// </summary>
		public RichTextView ()
		{
			this.Build ();
			_scrolledWindow = new GaplessScrolledWindow();
			_textView = new WebView();
			_textView.NavigationRequested += Handle_TextViewNavigationRequested;
			_textView.LoadFinished += Handle_TextViewLoadFinished;
			_scrolledWindow.Add(_textView);
			_scrolledWindow.Vadjustment.ValueChanged += HandleScrollWindowsVadjustmentValueChanged;
			this.Add(_scrolledWindow);
			_waitForLoad = new ManualResetEvent(true);
		}

		void Handle_TextViewLoadFinished (object o, LoadFinishedArgs args)
		{
			_loading = false;
			_waitForLoad.Set ();
		}

		void Handle_TextViewNavigationRequested (object o, NavigationRequestedArgs args)
		{
			var target = args.Request.Uri;
			if(target.StartsWith("about"))
			{
				//scroll to verse
				args.RetVal = 0;
				return;
			}
			
			//get versification
			var requestVersificationEventArgs = new RequestVersificationEventArgs();
			RequestVersification.Fire (this, requestVersificationEventArgs);
			var versification = requestVersificationEventArgs.Versification;
			
			var targetParts = target.Split (',');
			
			int mouseButton = 1;
			if(target.Contains (','))
			{
				mouseButton = int.Parse (targetParts[1]) + 1;
			}
			
			IVerseIdentifier verseTarget;
			if(versification.TryCreateVerseIdentifier(targetParts[0], out verseTarget))
			{
				
				LinkClickedEventArgs<IVerseIdentifier> linkClickedEventArgs = new LinkClickedEventArgs<IVerseIdentifier>(verseTarget, mouseButton);
				VerseLinkClicked.Fire(this, linkClickedEventArgs);
				args.RetVal = 1;
				return;
			}
			
			ActionLinkClicked.Fire(this, new LinkClickedEventArgs<string>(targetParts[0], mouseButton));
			args.RetVal = 1;
		}
		
		public event EventHandler<RequestVersificationEventArgs> RequestVersification;
		private IVersification _versification;
		
		private IVersification Versification
		{
			get
			{
				if(_versification == null)
				{
					var eventArgs = new RequestVersificationEventArgs();
					RequestVersification.Fire(this, eventArgs);
					_versification = eventArgs.Versification;
				}
				return _versification;
			}
		}
		
		public bool VScrollbarVisible
		{
			get
			{
				return _scrolledWindow.VScrollbar.Visible;	
			}
		}
		
		public bool Editable
		{
			get
			{
				return _textView.Editable;	
			}
			set
			{
				_textView.Editable = value;	
			}
		}
		
		/// <summary>
		/// Handles the scroll windows vadjustment value changed.
		/// Determines if the scroll bar is at the bottom and requests more
		/// results if it is.
		/// </summary>
		private void HandleScrollWindowsVadjustmentValueChanged (object sender, EventArgs args)
		{
			if(_scrolledWindow.Vadjustment.Upper - _scrolledWindow.Vadjustment.PageSize == _scrolledWindow.Vadjustment.Value)
			{
				ScrollbarAtBottom.Fire(this, EventArgs.Empty);
			}
			
			System.Action action = _adjustmentChangedAction;
			if(action != null)
			{
				Application.Invoke((_,__) =>
				{
					action();
					_adjustmentChangedAction = null;
				});
			}
		}

		
		/// <summary>
		/// Text including markup
		/// </summary>
		public string Text
		{
			set
			{
				_loading = true;
				_waitForLoad.Reset();
				_textView.LoadString (value, "text/html", "UTF-8", "");
				_adjustmentChangedAction = null;
			}
		}

		public void AppendText(string text)
		{
			text = text.Replace(Environment.NewLine, string.Empty);
			string appendScript =
				"var inner = document.getElementsByTagName(\"body\")[0].innerHTML;" +
				"document.getElementsByTagName(\"body\")[0].innerHTML = inner + \'" + text.Replace("'","&#39;") + "\';";
			
			ExecuteScript(appendScript);
		}
		
		public void ExecuteScript(string text)
		{
			if(_loading)
			{
				new Task(() => 
				{
					_waitForLoad.WaitOne();
					this.Invoke(() => _textView.ExecuteScript(text));
				}).Start();
				return;
			}
			
			_textView.ExecuteScript(text);
		}
		
		/// <summary>
		/// Occurs when a verse link is clicked.
		/// </summary>
		public event EventHandler<LinkClickedEventArgs<IVerseIdentifier>> VerseLinkClicked;
		
		/// <summary>
		/// Occurs when a action link is clicked.
		/// </summary>
		public event EventHandler<LinkClickedEventArgs<string>> ActionLinkClicked;
		
		public event EventHandler<EventArgs> ScrollbarAtBottom;
	}
}

